package com.vulcannis.icfp2002.game;

import java.util.*;
import com.vulcannis.icfp2002.util.*;

public class PackageManager
{
	public PackageManager( )
	{
		idMap = new HashMap( );
		knownLocationMap = new HashMap( );
		possibleLocationMap = new HashMap( );
		unreachablePoints = new HashSet( );
		ignoreLocations = new HashSet( );
	}
	
	public void update( final Point location, final PackageUpdate update )
	{
		final Set packages = update.getPackages( );
		
		// Update state of any packages thought to be here previously.
		Set previousPackages = (Set)knownLocationMap.get( location );
		if( previousPackages != null ) {
			Iterator it = previousPackages.iterator( );
			while( it.hasNext( ) ) {
				final Package pack = (Package)it.next( );
				pack.setLocation( null );
				pack.setLocationState( Package.UNKNOWN );
			}
			knownLocationMap.remove( location );
		}
		previousPackages = (Set)possibleLocationMap.get( location );
		if( previousPackages != null ) {
			Iterator it = previousPackages.iterator( );
			while( it.hasNext( ) ) {
				final Package pack = (Package)it.next( );
				pack.setLocation( null );
				pack.setLocationState( Package.UNKNOWN );
			}
			possibleLocationMap.remove( location );
		}
		
		// Update state for packages now known to be here.
		final Iterator it = packages.iterator( );
		while( it.hasNext( ) ) {
			final Package pack = (Package)it.next( );
			final Integer id = new Integer( pack.getId( ) );
			pack.setLocation( location );
			pack.setLocationState( Package.KNOWN_DROPPED );
			
			if( idMap.containsKey( id ) ) {
				final Package oldPack = (Package)idMap.get( id );
				removePackage( oldPack );
			}
			
			addPackage( pack );
		}
	}
	
	public void update( final PlayerUpdate update )
	{
		final PackageAction actions[ ] = update.getPackageActions( );
		for( int loop = 0; loop < actions.length; loop++ ) {
			final PackageCarrier player = actions[ loop ].getCarrier( );
			final int packageId = actions[ loop ].getPackageID( );
			final Point location = actions[ loop ].getLocation( );
			if( actions[ loop ].isPickup( ) )
				player.addPackage( updatePackage( packageId, location, Package.KNOWN_CARRIED ) );
			else {
				final Package pack = updatePackage( packageId, location, Package.POSSIBLE_DROPPED );
				if( pack != null ) {
					player.removePackage( pack );
				}
			}
		}
	}
	
	public void addUnreachablePoints( final Set points )
	{
		unreachablePoints.addAll( points );
		Iterator it = points.iterator( );
		while( it.hasNext( ) ) {
			final Point element = (Point)it.next( );
			Package pack = (Package)knownLocationMap.get( element );
			if( pack != null )
				removePackage( pack );
			pack = (Package)possibleLocationMap.get( element );
			if( pack != null )
				removePackage( pack );
		}
	}
	
	public Set getPackages( final Point loc )
	{
		return getPackages( loc, true );
	}
	
	public Set getPackages( final Point loc, final boolean screenForIgnores )
	{
		if( screenForIgnores && ignoreLocations.contains( loc ) )
			return Collections.EMPTY_SET;
		Set packages = (Set)knownLocationMap.get( loc );
		if( packages == null )
			return Collections.EMPTY_SET;
		else
			return packages;
	}
	
	public Set getKnownPackageLocations( )
	{
		return SetUtil.subtraction( knownLocationMap.keySet( ), ignoreLocations );
	}
	
	public Set getPossiblePackageLocations( )
	{
		return SetUtil.subtraction( possibleLocationMap.keySet( ), ignoreLocations );
	}
	
	public void addToIgnoreList( Set s )
	{
		ignoreLocations.addAll( s );
	}
	
	public void removeFromIgnoreList( Set s )
	{
		ignoreLocations.removeAll( s );
	}
	
	public void clearIgnoreList( )
	{
		ignoreLocations.clear( );
	}
	
	public void addToPossibleLocations( final Set locations )
	{
		final Iterator it = locations.iterator( );
		while( it.hasNext( ) ) {
			addPackage( new Package( (Point)it.next( ) ) );
		}
		removeFromIgnoreList( locations );
	}
	
	/**
	 * Updates (or creates if it doesn't exist yet) the state
	 * of the package given by id.
	 */
	private Package updatePackage( final int id, final Point location, final int state )
	{
		final Integer packageId = new Integer( id );
		Package pack = null;
		Point dest = null;
		int weight = -1;
		if( idMap.containsKey( packageId ) ) {
			pack = (Package)idMap.get( packageId );
			dest = pack.getDestination( );
			weight = pack.getWeight( );
			removePackage( pack );
		}
		if( state != Package.POSSIBLE_DROPPED || !location.equals( dest ) ) {
			pack = new Package( id, dest, weight );
			pack.setLocation( location );
			pack.setLocationState( state );
			addPackage( pack );
		}
		return pack;
	}
	
	private void addPackage( final Package pack )
	{
		final Point loc = pack.getLocation( );
		if( unreachablePoints.contains( loc ) )
			return;
		idMap.put( new Integer( pack.getId( ) ), pack );
		
		if( pack.getLocationState( ) == Package.POSSIBLE_DROPPED && !pack.getLocation( ).equals( pack.getDestination( ) ) )
			pack.setLocationState( Package.KNOWN_DROPPED );
			
//System.out.println( "Adding package: " + pack );
		if( pack.getLocationState( ) == Package.KNOWN_DROPPED )
			addLocationEntry( knownLocationMap, pack );
		else if( pack.getLocationState( ) == Package.POSSIBLE_DROPPED )
			addLocationEntry( possibleLocationMap, pack );
	}
	
	private void removePackage( final Package pack )
	{
//System.out.println( "Removing package: " + pack );
		idMap.remove( new Integer( pack.getId( ) ) );
		if( pack.getLocationState( ) == Package.KNOWN_DROPPED )
			removeLocationEntry( knownLocationMap, pack );
	}
	
	private static void addLocationEntry( final Map locationMap, final Package pack )
	{
//System.out.println( "Adding location entry: " + pack );
		final Point location = pack.getLocation( );
		Set packages = (Set)locationMap.get( location );
		if( packages == null )
			packages = new HashSet( );
		packages.add( pack );
		locationMap.put( location, packages );
	}
	
	private static void removeLocationEntry( final Map locationMap, final Package pack )
	{
//System.out.println( "Removing location entry: " + pack );
		final Point location = pack.getLocation( );
		final Set packages = (Set)locationMap.get( location );
		packages.remove( pack );
		if( packages.isEmpty( ) )
			locationMap.remove( location );
	}
	
	/**
	 * Maps package id's to the Package objects.
	 */
	private Map idMap;
	/**
	 * Maps locations to lists of packages known to be
	 * on the ground at that location.
	 */
	private Map knownLocationMap;
	/**
	 * Maps locations to the lists of packages that
	 * may possibly be at that point.
	 */
	private Map possibleLocationMap;
	private Set unreachablePoints;
	private Set ignoreLocations;
}
