package com.vulcannis.icfp2002.util;

import java.util.*;

final public class AStar
{
	public AStar( AStarMap map )
	{
		this.map = map;
		open = new NodeHeap( );
		inOpen = new byte[ map.getWidth( ) ][ map.getHeight( ) ];
		closed = new byte[ map.getWidth( ) ][ map.getHeight( ) ];
	}
	
	public Point[ ] getShortestPath( final Point start, final Destination dest )
	{
		return getShortestPath( start.getX( ), start.getY( ), dest );
	}
	
	private Point[ ] getShortestPath( final int startx, final int starty, final Destination dest )
	{
		reset( );
		
		final Node startNode = Node.newStartNode( startx, starty );
		final Node neighbours[ ] = new Node[ 4 ];
		
		open.push( startNode );
		setOpen( startNode.x, startNode.y, true );
		while( !open.isEmpty( ) ) {
			consideredCount++;
			final Node node = open.pop( );
			setOpen( node.x, node.y, false );
			
			if( dest.haveArrived( node.x, node.y ) ) {
				final Point path[ ] = buildPath( node );
				node.release( );
				return path;
			}
			
			final int numNeighbours = constructNeighbours( node, neighbours );
			for( int loop = 0; loop < numNeighbours; loop++ ) {
				final Node newNode = neighbours[ loop ];
				if( !inOpen( newNode.x, newNode.y ) && !inClosed( newNode.x, newNode.y ) ) {
					newNode.cost = dest.computeCost( newNode.x, newNode.y ) + node.costSoFar;
					open.push( newNode );
					setOpen( newNode.x, newNode.y, true );
				} else
					newNode.release( );
			}
			
			setClosed( node.x, node.y, true );
			node.release( );
		}
		return null;
	}
	
	public int getConsideredCount( )
	{
		return consideredCount;
	}
	
	public void dumpClosed( )
	{
		for( int y = map.getHeight( ); y > 0; y-- ) {
			for( int x = 1; x <= map.getWidth( ); x++ )
				System.out.print( inClosed( x, y )?'X':'-' );
			System.out.println( "" );
		}
	}
	
	private void reset( )
	{
		open.clear( );
		if( ++currentCallCount == 0 ) {
			currentCallCount++;
			final int width = map.getWidth( );
			final int height = map.getHeight( );
			for( int y = 0; y < height; y++ )
				for( int x = 0; x < width; x++ ) {
					inOpen[ x ][ y ] = closed[ x ][ y ] = 0;
				}
		}
		consideredCount = 0;
	}
	
	private boolean inOpen( int x, int y )
	{
		return inOpen[ x - 1 ][ y - 1 ] == currentCallCount;
	}
	
	private boolean inClosed( int x, int y )
	{
		return closed[ x - 1 ][ y - 1 ] == currentCallCount;
	}
	
	private void setOpen( int x, int y, boolean value )
	{
		if( value )
			inOpen[ x - 1 ][ y - 1 ] = currentCallCount;
		else
			inOpen[ x - 1 ][ y - 1 ] = (byte)( currentCallCount - 1 );
	}
	
	private void setClosed( int x, int y, boolean value )
	{
		if( value )
			closed[ x - 1 ][ y - 1 ] = currentCallCount;
		else
			closed[ x - 1 ][ y - 1 ] = (byte)( currentCallCount - 1 );
	}
	
	private Point[ ] buildPath( Node node )
	{
		final List list = new ArrayList( );
		while( node != null ) {
			list.add( new Point( node.x, node.y ) );
			node = node.parent;
		}
		final Point path[ ] = new Point[ list.size( ) ];
		final Iterator it = list.iterator( );
		int pos = path.length - 1;
		while( it.hasNext( ) )
			path[ pos-- ] = (Point)it.next( );
		return path;
	}
	
	private int constructNeighbours( Node node, Node neighbours[ ] )
	{
		int count = 0;
		if( map.isPassable( node.x + 1, node.y ) ) {
			neighbours[ count++ ] = Node.newChildNode( node.x + 1, node.y, node );
		}
		if( map.isPassable( node.x - 1, node.y ) ) {
			neighbours[ count++ ] = Node.newChildNode( node.x - 1, node.y, node );
		}
		if( map.isPassable( node.x, node.y + 1 ) ) {
			neighbours[ count++ ] = Node.newChildNode( node.x, node.y + 1, node );
		}
		if( map.isPassable( node.x, node.y - 1 ) ) {
			neighbours[ count++ ] = Node.newChildNode( node.x, node.y - 1, node );
		}
		for( int loop = count; loop < 4; loop++ ) {
			neighbours[ loop ] = null;
		}
		return count;
	}
	
	private AStarMap map;
	private NodeHeap open;
	private byte inOpen[ ][ ];
	private byte closed[ ][ ];
	private byte currentCallCount;
	private int consideredCount;
}

final class Node
{
	public int x, y;
	public Node parent;
	public int costSoFar;
	public int cost;

	public static Node newSentinel( )
	{
		final Node node = getNode( );
		return node;
	}
	
	public static Node newStartNode( final int x, final int y )
	{
		final Node node = getNode( );
		node.x = x;
		node.y = y;
		return node;
	}
	
	public static Node newChildNode( final int x, final int y, final Node parent )
	{
		final Node node = getNode( );
		node.x = x;
		node.y = y;
		node.parent = parent;
		if( parent == null )
			node.costSoFar = 0;
		else {
			node.costSoFar = parent.costSoFar + 1;
			parent.refCount++;
		}
		return node;
	}
	
	public void release( )
	{
		if( --refCount == 0 ) {
			x = y = cost = costSoFar = -1;
			if( parent != null )
				parent.release( );
			parent = null;
			if( unused.length == unusedCount ) {
				final Node newUnused[ ] = new Node[ unusedCount * 2 ];
				System.arraycopy( unused, 0, newUnused, 0, unusedCount );
				unused = newUnused;
			}
			unused[ unusedCount++ ] = this;
		}
	}
	
	public String toString( )
	{
		return "( " + x + ", " + y + " ): " + cost;
	}
	
	private Node( )
	{
		x = y = cost = costSoFar = -1;
		parent = null;
		refCount = 0;
	}
	
	private static Node getNode( )
	{
		Node node = null;
		if( unusedCount == 0 ) {
			node = new Node( );
		} else {
			node = unused[ --unusedCount ];
			unused[ unusedCount ] = null;
		}
		node.refCount++;
		return node;
	}
	
	private static Node unused[ ] = new Node[ 1024 ];
	private static int unusedCount;
	private int refCount;
}

final class NodeHeap
{
	public NodeHeap( )
	{
		data = new Node[ INITIAL_SIZE ];
		data[ 0 ] = Node.newSentinel( );
		count = 0;
	}
	
	public void push( Node node )
	{
		if( count == data.length - 1 ) {
			final Node newData[ ] = new Node[ data.length * 2 ];
			System.arraycopy( data, 0, newData, 0,  data.length );
			data = newData;
		}
		
		int pos = ++count;
		int parent = pos / 2;
		while( node.cost < data[ parent ].cost ) {
			data[ pos ] = data[ parent ];
			pos = parent;
			parent = pos / 2;
		}
		data[ pos ] = node;
	}
	
	public Node pop( )
	{
//		assert !isEmpty( );
		
		int pos = 1;
		Node min = data[ 1 ];
		Node last = data[ count-- ];
		for( int minChild = 0; pos * 2 <= count; pos = minChild ) {
			minChild = pos * 2;
			if( minChild != count && data[ minChild + 1 ].cost < data[ minChild ].cost )
				minChild++;
			if( last.cost > data[ minChild ].cost )
				data[ pos ] = data[ minChild ];
			else
				break;
		}
		data[ pos ] = last;
		
		return min;
	}
	
	public boolean isEmpty( )
	{
		return count == 0;
	}
	
	public void clear( )
	{
		for( int loop = 1; loop <= count; loop++ )
			data[ loop ].release( );
		count = 0;
	}
	
	private Node data[ ];
	private int count;
	
	private static final int INITIAL_SIZE = 1024;
}
